@@ -14,12 +14,12 @@ use sample_consensus::{Consensus, Estimator, Model};
 /// If you do not shuffle, the output will be biased towards data at the beginning
 /// of the inputs.
 pub struct Arrsac<R> {
+    initialization_hypotheses: usize,
+    initialization_blocks: usize,
     max_candidate_hypotheses: usize,
+    estimations_per_block: usize,
     block_size: usize,
-    max_delta_estimations: usize,
     likelihood_ratio_threshold: f32,
-    initial_epsilon: f32,
-    initial_delta: f32,
     inlier_threshold: f64,
     rng: R,
     random_samples: Vec<u32>,
@@ -37,23 +37,60 @@ where
     /// Some of the other parameters may need to be configured based on the amount of data,
     /// such as `block_size`, `likelihood_ratio_threshold`, and `block_size`. However,
     /// `inlier_threshold` has to be set based on the residual function used with the model.
+    ///
+    /// `initial_epsilon` must be higher than `initial_delta`. If you modify these values,
+    /// you need to make sure that within one `block_size` the `likelihood_ratio_threshold`
+    /// can be reached and a model can be rejected. Basically, make sure that
+    /// `((1.0 - delta) / (1.0 - epsilon))^block_size >>> likelihood_ratio_threshold`.
+    /// This must be done to ensure outlier models are rejected during the initial generation
+    /// phase, which only processes `block_size` datapoints.
+    ///
+    /// `initial_epsilon` should also be as large as you can set it where it is still relatively
+    /// pessimistic. This is so that we can more easily reject a model early in the process
+    /// to compute an updated value for delta during the adaptive process. This may not be possible
+    /// and will depend on your data.
     pub fn new(inlier_threshold: f64, rng: R) -> Self {
         Self {
-            max_candidate_hypotheses: 50,
-            block_size: 100,
-            max_delta_estimations: 4,
-            likelihood_ratio_threshold: 1e6,
-            initial_epsilon: 0.1,
-            initial_delta: 0.05,
+            initialization_hypotheses: 256,
+            initialization_blocks: 4,
+            max_candidate_hypotheses: 64,
+            estimations_per_block: 64,
+            block_size: 64,
+            likelihood_ratio_threshold: 1e3,
             inlier_threshold,
             rng,
             random_samples: vec![],
         }
     }
 
-    /// Number of hypotheses that will be generated for each block of data evaluated
+    /// Number of models generated in the initial step when epsilon and delta are being estimated.
     ///
-    /// Default: `50`
+    /// Default: `256`
+    pub fn initialization_hypotheses(self, initialization_hypotheses: usize) -> Self {
+        Self {
+            initialization_hypotheses,
+            ..self
+        }
+    }
+
+    /// Number of data blocks used to compute the initial estimate of delta and epsilon
+    /// before proceeding with regular block processing. This is used instead of
+    /// an initial epsilon and delta, which were suggested by the paper.
+    ///
+    /// Default: `4`
+    pub fn initialization_blocks(self, initialization_blocks: usize) -> Self {
+        Self {
+            initialization_blocks,
+            ..self
+        }
+    }
+
+    /// Maximum number of best hypotheses to retain during block processing
+    ///
+    /// This number is halved on each block such that on block `n` the number of
+    /// hypotheses retained is `max_candidate_hypotheses >> n`.
+    ///
+    /// Default: `64`
     pub fn max_candidate_hypotheses(self, max_candidate_hypotheses: usize) -> Self {
         Self {
             max_candidate_hypotheses,
@@ -61,28 +98,28 @@ where
         }
     }
 
-    /// Number of data points evaluated before more hypotheses are generated
+    /// Number of estmations (may generate multiple hypotheses) that will be ran
+    /// for each block of data evaluated
     ///
-    /// Default: `100`
-    pub fn block_size(self, block_size: usize) -> Self {
-        Self { block_size, ..self }
-    }
-
-    /// Number of times that the entire dataset is compared against a bad model to see
-    /// the probability of inliers in a bad model
-    ///
-    /// Default: `4`
-    pub fn max_delta_estimations(self, max_delta_estimations: usize) -> Self {
+    /// Default: `64`
+    pub fn estimations_per_block(self, estimations_per_block: usize) -> Self {
         Self {
-            max_delta_estimations,
+            estimations_per_block,
             ..self
         }
     }
 
+    /// Number of data points evaluated before more hypotheses are generated
+    ///
+    /// Default: `64`
+    pub fn block_size(self, block_size: usize) -> Self {
+        Self { block_size, ..self }
+    }
+
     /// Once a model reaches this level of unlikelihood, it is rejected. Set this
     /// higher to make it less restrictive, usually at the cost of more execution time.
     ///
-    /// Increasing this will make it more likely to find a good result (unless it is set very high).
+    /// Increasing this will make it more likely to find a good result.
     ///
     /// Decreasing this will speed up execution.
     ///
@@ -91,7 +128,7 @@ where
     /// **probability of rejecting a good model**. You can use that to control the probability
     /// that a good model is rejected.
     ///
-    /// Default: `1e6`
+    /// Default: `1e3`
     pub fn likelihood_ratio_threshold(self, likelihood_ratio_threshold: f32) -> Self {
         Self {
             likelihood_ratio_threshold,
@@ -99,36 +136,6 @@ where
         }
     }
 
-    /// Initial anticipated probability of an inlier being part of a good model
-    ///
-    /// This is an estimation that will be updated as ARRSAC executes. The initial
-    /// estimate is purposefully low, which will accept more models. As models are
-    /// accepted, it will gradually increase it to match the best model found so far,
-    /// which makes it more restrictive.
-    ///
-    /// Default: `0.1`
-    pub fn initial_epsilon(self, initial_epsilon: f32) -> Self {
-        Self {
-            initial_epsilon,
-            ..self
-        }
-    }
-
-    /// Initial anticipated probability of an inlier being part of a bad model
-    ///
-    /// This is an estimation that will be updated as ARRSAC executes. The initial
-    /// estimate is almost certainly incorrect. This can be modified for different data
-    /// to get better/faster results. As models are rejected, it will update this value
-    /// until it has evaluated it `max_delta_estimations` times.
-    ///
-    /// Default: `0.05`
-    pub fn initial_delta(self, initial_delta: f32) -> Self {
-        Self {
-            initial_delta,
-            ..self
-        }
-    }
-
     /// Residual threshold for determining if a data point is an inlier or an outlier of a model
     pub fn inlier_threshold(self, inlier_threshold: f64) -> Self {
         Self {
@@ -137,94 +144,87 @@ where
         }
     }
 
-    /// Algorithm 3 from "A Comparative Analysis of RANSAC Techniques Leading to Adaptive Real-Time Random Sample Consensus"
+    /// Adapted from algorithm 3 from "A Comparative Analysis of RANSAC Techniques Leading to Adaptive
+    /// Real-Time Random Sample Consensus", but it was effectively rewritten to avoid the need for
+    /// initial epsilon and delta.
     ///
-    /// At least at present, this does not use the PROSAC method and instead does completely random sampling.
-    ///
-    /// Returns the initial models (and their num inliers), `epsilon`, and `delta` in that order.
+    /// Returns the initial models (and their num inliers) sorted by decreasing inliers
+    /// and `delta` in that order.
     fn initial_hypotheses<E, Data>(
         &mut self,
         estimator: &E,
         data: impl Iterator<Item = Data> + Clone,
-    ) -> (Vec<(E::Model, usize)>, f32, f32)
+    ) -> (Vec<(E::Model, usize)>, f32)
     where
         E: Estimator<Data>,
     {
+        assert!(
+            self.initialization_blocks > 0,
+            "ARRSAC must have at least 1 initialization block"
+        );
+        // NOTE: This whole function is different than that specified in the ARRSAC paper.
+        // The reason is that you needed to provide a good initial guess for epsilon and delta
+        // otherwise it could lead to delta exceeding epsilon or situations where models could no
+        // longer be rejected or were almost always rejected. This solution is an imperfect compomise
+        // that assumes that delta will be roughly equal to the inlier ratio of the worst model generated,
+        // which both assumes that the worst model is an outlier and that it is actually representative of the
+        // population. The assumption of epsilon also assumes that the best model is an inlier, but is an
+        // otherwise good initial guess. The other caveat with this approach is that a sufficiently large
+        // set of initial datapoints is required to be able to accurately determine epsilon and delta.
+        // Therefore a new paremeter is added to separate the normal blocks from the initial generation set.
         let mut hypotheses = vec![];
         // We don't want more than `block_size` data points to be used to evaluate models initially.
-        let initial_datapoints = core::cmp::min(self.block_size, data.clone().count());
-        // Set the best inliers to be the floor of what the number of inliers would need to be to be the initial epsilon.
-        let mut best_inliers = (self.initial_epsilon * initial_datapoints as f32) as usize;
-        // Set the initial epsilon (inlier ratio in good model).
-        let mut epsilon = self.initial_epsilon;
-        // Set the initial delta (outlier ratio in good model).
-        let mut delta = self.initial_delta;
-        let mut positive_likelihood_ratio = delta / epsilon;
-        let mut negative_likelihood_ratio = (1.0 - delta) / (1.0 - epsilon);
-        let mut current_delta_estimations = 0;
-        let mut total_delta_inliers = 0;
-        let mut best_inlier_indices = vec![];
-        let mut random_hypotheses = vec![];
-        // Lets us know if we found a candidate hypothesis that actually has enough inliers for us to generate a model from.
-        let mut found_usable_hypothesis = false;
-        // Iterate through all the randomly generated hypotheses to update epsilon and delta while finding good models.
-        for _ in 0..self.max_candidate_hypotheses {
-            if found_usable_hypothesis {
-                // If we have found a hypothesis that has a sufficient number of inliers, we randomly sample from its inliers
-                // to generate new hypotheses since that is much more likely to generate good ones.
-                random_hypotheses.extend(self.generate_random_hypotheses_subset(
-                    estimator,
-                    data.clone(),
-                    &best_inlier_indices,
-                ));
-            } else {
-                // Generate the random hypotheses using all the data, not just the evaluation data.
-                random_hypotheses.extend(self.generate_random_hypotheses(estimator, data.clone()));
-            }
-            for model in random_hypotheses.drain(..) {
-                // Check if the model satisfies the ASPRT test on only `inital_datapoints` evaluation data.
-                let (pass, inliers) = self.asprt(
-                    data.clone().take(initial_datapoints),
-                    &model,
-                    positive_likelihood_ratio,
-                    negative_likelihood_ratio,
-                    E::MIN_SAMPLES,
-                );
-                if pass {
-                    // If this has the largest support (most inliers) then we update the
-                    // approximation of epsilon.
-                    if inliers > best_inliers {
-                        best_inliers = inliers;
-                        // Update epsilon (this can only increase, since there are more inliers).
-                        epsilon = inliers as f32 / initial_datapoints as f32;
-                        // Will decrease positive likelihood ratio.
-                        positive_likelihood_ratio = delta / epsilon;
-                        // Will increase negative likelihood ratio.
-                        negative_likelihood_ratio = (1.0 - delta) / (1.0 - epsilon);
-
-                        // Update the inlier indices appropriately.
-                        best_inlier_indices = self.inliers(data.clone(), &model);
-                        // Mark that a usable hypothesis has been found.
-                        found_usable_hypothesis = true;
-                    }
-                    hypotheses.push((model, inliers));
-                } else if current_delta_estimations < self.max_delta_estimations {
-                    // We want to add the information about inliers to our estimation of delta.
-                    // We only do this up to `max_delta_estimations` times to avoid wasting too much time.
-                    total_delta_inliers += self.count_inliers(data.clone(), &model);
-                    current_delta_estimations += 1;
-                    // Update delta.
-                    delta = total_delta_inliers as f32
-                        / (current_delta_estimations * data.clone().count()) as f32;
-                    // May change positive likelihood ratio.
-                    positive_likelihood_ratio = delta / epsilon;
-                    // May change negative likelihood ratio.
-                    negative_likelihood_ratio = (1.0 - delta) / (1.0 - epsilon);
-                }
+        let initial_datapoints = core::cmp::min(
+            self.initialization_blocks * self.block_size,
+            data.clone().count(),
+        );
+        // Generate the initial batch of random hypotheses and count their inliers and outliers.
+        for _ in 0..self.initialization_hypotheses {
+            for model in self.generate_random_hypotheses(estimator, data.clone()) {
+                let inliers = self.count_inliers(data.clone().take(initial_datapoints), &model);
+                hypotheses.push((model, inliers));
             }
         }
 
-        (hypotheses, epsilon, delta)
+        // Sort the hypotheses by their inliers.
+        hypotheses.sort_unstable_by_key(|&(_, inliers)| Reverse(inliers));
+
+        // Compute epsilon and delta using the best and worst model generated.
+        let epsilon = hypotheses
+            .first()
+            .map(|&(_, inliers)| inliers as f32 / initial_datapoints as f32)
+            .unwrap_or_default();
+        let delta = hypotheses
+            .last()
+            .map(|&(_, inliers)| if inliers < E::MIN_SAMPLES {E::MIN_SAMPLES} else {inliers} as f32 / initial_datapoints as f32)
+            .unwrap_or_default();
+
+        if epsilon < delta {
+            // If epsilon is less than delta, then better hypotheses will get rejected and worse accepted,
+            // which is counter to what we want. In this case, we had a bad initialization, so clear the hypotheses.
+            // This will cause execution to terminate.
+            hypotheses.clear();
+            return (hypotheses, delta);
+        }
+
+        // Populate hypotheses with hypotheses generated from the inliers of the best hypothesis.
+        // This will use the initialization datapoints and filter with SPRT.
+        self.populate_hypotheses_sprt(
+            estimator,
+            &mut hypotheses,
+            delta,
+            data,
+            initial_datapoints,
+            self.initialization_hypotheses,
+        );
+
+        // Sort the hypotheses by their inliers.
+        hypotheses.sort_unstable_by_key(|&(_, inliers)| Reverse(inliers));
+
+        // Filter down the hypotheses to just the best ones.
+        hypotheses.truncate(self.max_candidate_hypotheses >> (self.initialization_blocks - 1));
+
+        (hypotheses, delta)
     }
 
     /// Populates `self.random_samples` using a len.
@@ -251,6 +251,55 @@ where
         }
     }
 
+    fn populate_hypotheses_sprt<E, Data>(
+        &mut self,
+        estimator: &E,
+        hypotheses: &mut Vec<(E::Model, usize)>,
+        delta: f32,
+        data: impl Iterator<Item = Data> + Clone,
+        num_checked: usize,
+        num_hypotheses: usize,
+    ) where
+        E: Estimator<Data>,
+    {
+        // Update epsilon using the best model.
+        // Since epsilon can only increase and delta is fixed, we can be sure that these ratios
+        // will still be valid (epsilon > delta).
+        let epsilon = hypotheses[0].1 as f32 / num_checked as f32;
+        // Create the likelihood ratios for inliers and outliers.
+        let positive_likelihood_ratio = delta / epsilon;
+        let negative_likelihood_ratio = (1.0 - delta) / (1.0 - epsilon);
+        // Generate the list of inliers for the best model.
+        let mut inliers = self.inliers(data.clone().take(num_checked), &hypotheses[0].0);
+        if inliers.len() <= E::MIN_SAMPLES {
+            // If we don't have enough samples to generate more models, then we should expand the inliers to
+            // the entire dataset.
+            inliers = self.inliers(data.clone().take(num_checked), &hypotheses[0].0);
+        }
+        // We generate hypotheses until we reach the initial num hypotheses.
+        // We can't count the number generated because it could generate 0 hypotheses
+        // and then the loop would continue indefinitely.
+        let mut random_hypotheses = Vec::new();
+        for _ in 0..num_hypotheses {
+            random_hypotheses.extend(self.generate_random_hypotheses_subset(
+                estimator,
+                data.clone(),
+                &inliers,
+            ));
+            for model in random_hypotheses.drain(..) {
+                if let Some(inliers) = self.asprt(
+                    data.clone().take(num_checked),
+                    &model,
+                    positive_likelihood_ratio,
+                    negative_likelihood_ratio,
+                    E::MIN_SAMPLES,
+                ) {
+                    hypotheses.push((model, inliers));
+                }
+            }
+        }
+    }
+
     /// Generates as many hypotheses as one call to `Estimator::estimate()` returns from all data.
     fn generate_random_hypotheses<E, Data>(
         &mut self,
@@ -288,7 +337,7 @@ where
 
     /// Algorithm 1 in "Randomized RANSAC with Sequential Probability Ratio Test".
     ///
-    /// This tests if a model is accepted. Returns `true` on accepted and `false` on rejected.
+    /// This tests if a model is accepted. Returns `Some(inliers)` if accepted or `None` if rejected.
     ///
     /// `inlier_threshold` - The model residual error threshold between inliers and outliers
     /// `positive_likelihood_ratio` - `δ / ε`
@@ -300,7 +349,7 @@ where
         positive_likelihood_ratio: f32,
         negative_likelihood_ratio: f32,
         minimum_samples: usize,
-    ) -> (bool, usize) {
+    ) -> Option<usize> {
         let mut likelihood_ratio = 1.0;
         let mut inliers = 0;
         for data in data {
@@ -311,12 +360,12 @@ where
                 negative_likelihood_ratio
             };
 
-            if likelihood_ratio > self.likelihood_ratio_threshold {
-                return (false, 0);
+            if likelihood_ratio > self.likelihood_ratio_threshold || likelihood_ratio.is_nan() {
+                return None;
             }
         }
 
-        (inliers >= minimum_samples, inliers)
+        (inliers >= minimum_samples).then(|| inliers)
     }
 
     /// Determines the number of inliers a model has.
@@ -364,26 +413,17 @@ where
         if data.clone().count() < E::MIN_SAMPLES {
             return None;
         }
-        // Generate the initial set of hypotheses. This also gets us an estimate of epsilon and delta.
-        // We only want to give it one block size of data for the initial generation.
-        let (mut hypotheses, _, delta) = self.initial_hypotheses(estimator, data.clone());
+        // Generate the initial set of hypotheses. This also gets us an estimate of delta.
+        let (mut hypotheses, delta) = self.initial_hypotheses(estimator, data.clone());
 
-        let mut random_hypotheses = Vec::new();
-
-        // Retain the hypotheses the initial time. This is done before the loop to ensure that if the
-        // number of datapoints is too low and the for loop never executes that the best model is returned.
-        hypotheses.sort_unstable_by_key(|&(_, inliers)| Reverse(inliers));
-        hypotheses.truncate(self.max_candidate_hypotheses);
-
-        // If there are no initial hypotheses or the best hypothesis doesnt have enough inliers then don't bother doing anything.
-        if hypotheses.is_empty()
-            || self.inliers(data.clone(), &hypotheses[0].0).len() <= E::MIN_SAMPLES
-        {
+        // If there are no initial hypotheses then initialization failed, so exit early.
+        if hypotheses.is_empty() {
             return None;
         }
 
         // Gradually increase how many datapoints we are evaluating until we evaluate them all.
-        'outer: for block in 1.. {
+        // This starts at the first block that was not evaluated in initial_hypotheses.
+        'outer: for block in self.initialization_blocks.. {
             let samples_up_to_beginning_of_block = block * self.block_size;
             let samples_up_to_end_of_block = samples_up_to_beginning_of_block + self.block_size;
             // Score hypotheses with samples.
@@ -401,45 +441,23 @@ where
                     }
                 }
             }
-            // First, update epsilon using the best model.
-            // Technically model 0 might no longer be the best model after evaluating the last data-point,
-            // but that is not that important.
-            let epsilon = hypotheses[0].1 as f32 / samples_up_to_end_of_block as f32;
-            // Create the likelihood ratios for inliers and outliers.
-            let positive_likelihood_ratio = delta / epsilon;
-            let negative_likelihood_ratio = (1.0 - delta) / (1.0 - epsilon);
-            // Generate the list of inliers for the best model.
-            let inliers = self.inliers(data.clone(), &hypotheses[0].0);
-            // We generate hypotheses until we reach the initial num hypotheses.
-            // We can't count the number generated because it could generate 0 hypotheses
-            // and then the loop would continue indefinitely.
-            for _ in 0..self.max_candidate_hypotheses {
-                random_hypotheses.extend(self.generate_random_hypotheses_subset(
-                    estimator,
-                    data.clone(),
-                    &inliers,
-                ));
-                for model in random_hypotheses.drain(..) {
-                    let (pass, inliers) = self.asprt(
-                        data.clone().take(samples_up_to_end_of_block),
-                        &model,
-                        positive_likelihood_ratio,
-                        negative_likelihood_ratio,
-                        E::MIN_SAMPLES,
-                    );
-                    if pass {
-                        hypotheses.push((model, inliers));
-                    }
-                }
-            }
+            // Sort the hypotheses by their inliers to find the best.
+            hypotheses.sort_unstable_by_key(|&(_, inliers)| Reverse(inliers));
+            // Populate hypotheses with hypotheses that pass SPRT.
+            self.populate_hypotheses_sprt(
+                estimator,
+                &mut hypotheses,
+                delta,
+                data.clone(),
+                samples_up_to_end_of_block,
+                self.estimations_per_block,
+            );
             // This will retain at least half of the hypotheses each time
             // and gradually decrease as the number of samples we are evaluating increases.
-            // NOTE: The paper says to run this outside of this if statement, but that
-            // seems incorrect intuitively, other wise the minimum would cause the
-            // number of models to halve on every single data point.
-            // At least halving on every block makes more sense.
-            // The paper also says to use a peculiar formula that just results in doing
-            // this basic right shift below.
+            // NOTE:
+            // The paper says to use a peculiar formula that just results in doing
+            // this basic right shift below, but as written it contained some apparent errors in
+            // where it was ran. This seems to be the correct location to do this.
             hypotheses.sort_unstable_by_key(|&(_, inliers)| Reverse(inliers));
             hypotheses.truncate(self.max_candidate_hypotheses >> block);
             if hypotheses.len() <= 1 {
